import * as v from 'valibot';

interface BaseEndpointState {
	operations: unknown;
}

type OpValueType = undefined | boolean;

export interface Endpoint<Schema extends {} = {}, State extends BaseEndpointState = BaseEndpointState> {
	schema: Schema;
	can<OpName extends string = string, OpValue extends OpValueType = OpValueType>(
		op: OpName,
		value?: OpValue
	): Endpoint<
		Schema,
		{
			operations: State['operations'] extends string[] ? readonly [...State['operations'], OpName] : [OpName],
		}
	>;
}

interface EndpointParams<Schema extends v.BaseSchema = v.BaseSchema> {
	schema: Schema;
}

class EndpointInstance<
	Params extends EndpointParams,
	State extends BaseEndpointState
> implements Endpoint<Params['schema'], State> {
	readonly operations: Set<string>;
	readonly schema: Params['schema'];

	constructor(params: Params) {
		this.schema = params.schema;
		this.operations = new Set<string>();
	}

	can<OpName extends string = string, OpValue extends OpValueType = OpValueType>(
		op: OpName,
		value?: OpValue
	): Endpoint<
		Params['schema'],
		{
			operations: State['operations'] extends string[] ? readonly [...State['operations'], OpName] : [OpName],
		}
	> {
		if (value) {
			this.operations.add(op);
		} else {
			this.operations.delete(op);
		}

		return this;
	}
}

export const endpoint = <Params extends EndpointParams>(params: Params): Endpoint<v.Output<Params['schema']>> => {
	return new EndpointInstance(params);
};

export interface BaseAppState {
	endpoints: unknown;
	operations: unknown;
}

type EndpointOperations<T extends Endpoint> = T extends Endpoint<any, infer R> ? (
	R extends { operations: Record<number, any> } ? R['operations'][number] : []
	) : [];

type AppOperations<T extends App> = (
	T extends App<any, infer R>
		? R extends BaseAppState
			? keyof R['operations']
			: never
		: never
	);



export interface App<Name extends string = string, State extends BaseAppState = {
	endpoints: [];
	operations: Record<never, []>;
}> {
	name: Name;
	operation<
		OperationName extends string,
		OperationParams extends BaseOperationParams<OperationName>,
		NewOperation extends Operation<OperationParams>
	>(newOperation: NewOperation): App<
		Name,
		{
			endpoints: State['endpoints'],
			operations: keyof State['operations'] extends never ? {
				[Key in NewOperation['name']]: (
					Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
					)
			} : {
				[Key in NewOperation['name'] | keyof State['operations']]: (
					State['operations'] extends Record<Key, any>
						? (
							State['operations'][Key] extends readonly string[] ? State['operations'][Key] : (
								Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
								)
							)
						: (
							Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
							)
					);
			}
		}
	>;
	endpoint<NewEndpoint extends Endpoint = Endpoint>(newEndpoint: EndpointOperations<NewEndpoint> extends AppOperations<this> ? NewEndpoint : never): App<
		Name,
		{
			endpoints: State['endpoints'] extends Array<unknown> ? [...State['endpoints'], NewEndpoint] : [NewEndpoint],
			operations: State['operations']
		}
	>;
}

interface BaseOperationParams<Name extends string = string, Args extends readonly string[] = readonly string[]> {
	name: Name;
	args?: Args;
}

interface Operation<Params extends BaseOperationParams = BaseOperationParams> {
	name: Params['name'];
	args: Params['args'];
}

class OperationInstance<Params extends BaseOperationParams = BaseOperationParams> implements Operation<Params> {
	readonly name: Params['name'];
	readonly args: Params['args'];

	constructor(params: Params) {
		this.name = params.name;
		this.args = params.args;
	}
}

export const operation = <Params extends BaseOperationParams = BaseOperationParams>(
	params: Params
): Operation<Params> => {
	return new OperationInstance(params);
};

interface AppParams<Name extends string = string> {
	name: Name;
}

class AppInstance<Params extends AppParams, State extends BaseAppState> implements App<Params['name'], State> {
	readonly name: Params['name'];
	readonly endpoints: Set<Endpoint>;
	readonly operations: Set<Operation>;

	constructor(params: Params) {
		this.name = params.name;
		this.endpoints = new Set<Endpoint>();
		this.operations = new Set<Operation>();
	}

	operation<NewOperation extends Operation>(newOperation: NewOperation): App<
		Params['name'],
		{
			endpoints: State['endpoints'],
			operations: keyof State['operations'] extends never ? {
				[Key in NewOperation['name']]: (
					Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
				)
			} : {
				[Key in NewOperation['name'] | keyof State['operations']]: (
					State['operations'] extends Record<Key, any>
						? (
							State['operations'][Key] extends readonly string[] ? State['operations'][Key] : (
								Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
								)
							)
						: (
							Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
							)
					);
			}
		}
	> {
		this.operations.add(newOperation);
		return this as App<
			Params['name'],
			{
				endpoints: State['endpoints'],
				operations: {
					[Key in NewOperation['name'] | keyof State['operations']]: (
						State['operations'] extends Record<Key, any>
							? (
								State['operations'][Key] extends readonly string[] ? State['operations'][Key] : (
									Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
									)
								)
							: (
								Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
								)
						);
				}
			}
		>;
	}

	endpoint<NewEndpoint extends Endpoint = Endpoint>(newEndpoint: NewEndpoint): App<
		Params['name'],
		{
			endpoints: State['endpoints'] extends Array<unknown> ? [...State['endpoints'], NewEndpoint] : [NewEndpoint],
			operations: State['operations']
		}
	> {
		this.endpoints.add(newEndpoint);
		return this;
	}
}

export const app = <Params extends AppParams>(params: Params): App<Params['name']> => {
	return new AppInstance(params);
};

export * as validation from 'valibot';
